#define G_INLINE_FUNC
#include <X11/Xlib.h>
#include <gtk/gtk.h>
#include <stdio.h>

static float rwxold, rwyold;

#define FACTZ  0.03          /* scaling factor */

#ifndef TRUE
#define TRUE   1             /* numeric value for true  */
#endif
#ifndef FALSE
#define FALSE  0             /* numeric value for false */
#endif

#define LWDATR TRUE          /* line width  capability */
#define LCLATR TRUE          /* line color  capability */
#define LTNATR TRUE          /* hard fill   capability */
#define LIMATR TRUE          /* bit image   capability */
#define LPTATR FALSE         /* mouse point capability */

#define MAXWDI 9             /* maximum number of line index */
#define MAXCLI 255           /* maximum number of line color */
#define MAXBMP 300           /* maximum number of bitmaps */

static int linewidth[MAXWDI] = { 1, 2, 2, 3, 3, 4, 4, 5, 5 };

static int iwdidz;           /* pen width */
static int iclidz;           /* pen color index */
static int iwtroz;           /* direction of frame */
static int lclatrz;          /* type of screen */

static int ixz, iyz, iwz, ihz, ixxz, iyyz; /* for image */

static int nbmap, nn1[MAXBMP], nn2[MAXBMP], nx[MAXBMP], ny[MAXBMP];
static char bmline[MAXBMP][260];

static int lfirst = TRUE;

static GdkGC *gc;
static GdkGC *gct;
static GtkWidget *drawing_area;
static GdkPixmap *pixmap;

static  GdkColor cx[MAXCLI];

/*------------------------- GDK  ------------------------*/

zgsdrw_(d)
     GtkWidget *d;
{

  /* set/reset drawing_area */

  drawing_area = d;
}

zgspmp_(pmap)
     GdkPixmap *pmap;
{

  /* set/reset pixmap */

  pixmap = pmap;
}

/*------------------------- device ------------------------*/

void zgdopn_(clrmap, cbmmap)
    char clrmap[], cbmmap[];
{
  int ncolor, n, m;

  guint16 rx[MAXCLI], gx[MAXCLI], bx[MAXCLI];
  char c[80], cmapz[80], bmapz[80];

  FILE *stream;

  GdkVisual *vis;
  GdkColormap *cmap;

  /* check drawing_area, if not set, stop */

  if (drawing_area == NULL) {
    fprintf(stderr, "*** Error in zgdopn: ");
    fprintf(stderr,
	    "drawable is not setted.\n");
    exit(1);
  }

  /* read colormap file */

  cfnchr(cmapz, clrmap, 79);

  if ((stream = fopen(cmapz, "r")) == NULL) {
    fprintf(stderr, "*** Error in zgdopn : ");
    fprintf(stderr,
	    "Allocation failed for colormap (%s).\n", cmapz);
    exit (1);
  }

  fscanf(stream, "%d : %s", &ncolor, c);
  for (n = 0; n < ncolor; n++)
    fscanf(stream, "%6hd%6hd%6hd : %s", &rx[n], &gx[n], &bx[n], c);

  /* read bitmap file */

  cfnchr (bmapz, cbmmap, 79);

  if ((stream = fopen(bmapz, "r")) == NULL) {
    fprintf(stderr, "*** Error in zgdopn : ");
    fprintf(stderr,
	     "Allocation failed for bitmap (%s).\n", bmapz);
    exit (1);
  }

  fscanf(stream, "%d", &nbmap);
  for (n = 0; n < nbmap; n++)
    fscanf(stream, "%4d%4d%3d%3d%s",
	   &nn1[n], &nn2[n], &nx[n], &ny[n], bmline[n]);

  /* set colormap */

  vis = gtk_widget_get_visual(drawing_area);

  lclatrz = (vis->type != GDK_VISUAL_STATIC_GRAY);

  if (lclatrz) {

    cmap = gtk_widget_get_colormap(drawing_area);

    for (n = 0; n < MAXCLI; n++) {

      m = n % ncolor;

      cx[n].red   = rx[m];
      cx[n].green = gx[m];
      cx[n].blue  = bx[m];

      if (!gdk_colormap_alloc_color(cmap, &cx[n], FALSE, FALSE))
	break;

    }

    if (n < MAXCLI) {

      fprintf (stderr, "*** Warning in zgdopn : ");
      fprintf (stderr, "Only %d colors are allocated.\n", n);

      for ( ; n < MAXCLI; n++)
	cx[n].pixel = drawing_area->style->black.pixel;

    }

  } else {
    cx[0].pixel = drawing_area->style->white.pixel;
    for (n = 1; n < MAXCLI; n++)
      cx[n].pixel = drawing_area->style->black.pixel;
  }

  gc = gdk_gc_new(drawing_area->window);
}

void zgdcls_()
{

  /* device closing proc. */

}

/*------------------------- page --------------------------*/

void zgpopn_()
{

  /* background drawing proc. */

}

void zgpcls_()
{

  /* event loop */

}

/*------------------------- object ------------------------*/

void zgoopn_(objname, comment)
    char *objname, *comment;
{

  /* for long-type message? */

}

void zgocls_(objname)
    char *objname;
{

  /* for long-type message? */

}

/*------------------------- line --------------------------*/

void zgswdi_(iwdidx)
    int *iwdidx;
{

  /* set line width index */

  iwdidz = *iwdidx % 10;

  if (iwdidz == 0)
    iwdidz = 1;

}

void zgscli_(iclidx)
    int *iclidx;
{

  /* set line color index */

  iclidz = *iclidx % MAXCLI;

  if (iclidz == 0)
    iclidz = 1;
}

void zggopn_()
{

  /* open graphic segment */

  gdk_gc_set_foreground(gc,
			&cx[iclidz]);

  gdk_gc_set_line_attributes(gc,
			     linewidth[iwdidz-1],
			     GDK_LINE_SOLID,
			     GDK_CAP_ROUND,
			     GDK_JOIN_BEVEL);
}

void zggmov_(wx, wy)
     float *wx, *wy;
{

  /* pen-up move */

  rwxold = *wx;
  rwyold = *wy;

}

void zggplt_(wx, wy)
     float *wx, *wy;
{
  int iwxold, iwyold, iwx, iwy;
  void zgfint_();

  /* pen-down move */

  zgfint_(&rwxold, &rwyold, &iwxold, &iwyold);
  zgfint_( wx,      wy,     &iwx,    &iwy   );

  if (iwxold == iwx && iwyold == iwy)
    gdk_draw_point(pixmap, gc, iwx, iwy);
  else
    gdk_draw_line(pixmap, gc, iwxold, iwyold, iwx, iwy);

  rwxold = *wx;
  rwyold = *wy;
}

void zggcls_()
{

  /* close graphic segment */

}

/*------------------------- tone --------------------------*/

void zggton_(np, wpx, wpy, itpat)
     int *np, *itpat;
     float wpx[], wpy[];
{
  static int ltfrst = TRUE, ibitold = -1;
  static GdkPoint p[256];
  int ipx, ipy, iclr, ibit;
  int i, nb;
  char bitmap[256];

  GdkPixmap *pixmap;

  void zgbmcv_(), zgfint_();

  /* hard fill */

  if (ltfrst) {
    gct = gdk_gc_new (drawing_area->window);

    ltfrst = FALSE;
  }

  ibit = *itpat % 1000;
  iclr = (*itpat / 1000) % MAXCLI;
  if (iclr == 0)
    iclr = 1;

  gdk_gc_set_foreground (gct, &cx[iclidz]);

  if (ibit != ibitold) {
    ibitold = ibit;
    nb = -1;
    if (iwtroz == 1) {
      for (i = 0; i < nbmap; i++)
	if (ibit == nn1[i]) {
	  nb = i;
	  break;
	}
    }
    else {
      for (i = 0; i < nbmap; i++)
	if (ibit == nn2[i]) {
	  nb = i;
	  break;
	}
    }
    if (nb >= 0) {
      zgbmcv_(&nx[nb], &ny[nb], bmline[nb], bitmap);
      pixmap = gdk_bitmap_create_from_data(NULL, bitmap, nx[nb], ny[nb]);
      gdk_gc_set_fill(gct, GDK_STIPPLED);
      gdk_gc_set_stipple(gct, pixmap);
    }
    else
      return;
  }

  for (i = 0; i < *np; i++) {
    zgfint_(&wpx[i],&wpy[i], &ipx, &ipy);
    p[i].x = ipx;
    p[i].y = ipy;
  }
  gdk_draw_polygon(pixmap, gct, TRUE, p, *np);
}

/*------------------------- image -------------------------*/

void zgiopn_(iwx, iwy, iwidth, iheight)
     int *iwx, *iwy, *iwidth, *iheight;
{
  ixz = *iwx;
  iyz = *iwy;
  iwz = *iwidth;
  ihz = *iheight;
  ixxz = ixz;
  iyyz = iyz;
  printf (" *** image ");
}

void zgidat_(image, nlen)
     int *nlen, image[];
{
  int i;

  for (i = 0; i < *nlen; i++) {
    gdk_gc_set_foreground(gc, &cx[image[i]]);
    gdk_draw_point(pixmap, gc, ixxz, iyyz);
    ixxz = ixxz + 1;
    if (ixxz >= ixz+iwz) {
      ixxz = ixz;
      iyyz = iyyz + 1;
      if (iyyz % 16 == 0) {
	printf (".");
	fflush (stdout);
      }
    }
  }
}

void zgicls_()
{
    printf (" end\n");
}

/*------------------------- mouse -------------------------*/

void zgqpnt_(wx, wy, mb)
     float *wx, *wy;
     int *mb;
{

  /* mouse pointer position info. */

}

/*---------------------- transformation -------------------*/

void zgfint_(wx, wy, iwx, iwy)
     float *wx, *wy;
     int *iwx, *iwy;
{
  int dheight;
  int dwidth;

  *iwx = *wx + 0.5;
  *iwy = drawing_area->allocation.height - *wy + 0.5;
}

void zgiint_(iwx, iwy, wx, wy)
     float *wx, *wy;
     int *iwx, *iwy;
{
  *wx = *iwx;
  *wy = drawing_area->allocation.height - *iwy;
}

/*------------------------- inquiry -----------------------*/

void zgqwdc_(lwdatr)
     int *lwdatr;
{

  /* inquire line width capability */

  *lwdatr = LWDATR;
}

void zgqclc_(lclatr)
     int *lclatr;
{

  /* inquire line color capability */

  *lclatr = lclatrz;
}

void zgqtnc_(ltnatr)
     int *ltnatr;
{

  /* inquire hard fill capability */

  *ltnatr = LTNATR;
}

void zgqimc_(limatr)
     int *limatr;
{

  /* inquire bit image capability */

  *limatr = LIMATR;
}

void zgqptc_(lptatr)
     int *lptatr;
{

  /* inquire mouse point capability */

  *lptatr = LPTATR;
}

void zgqrct_(wsxmn, wsxmx, wsymn, wsymx, fact)
     float *wsxmn, *wsxmx, *wsymn, *wsymx, *fact;
{

  /* inquire workstation rectangle */

  *wsxmn = 0.;
  *wsxmx = drawing_area->allocation.width;
  *wsymn = 0.;
  *wsymx = drawing_area->allocation.height;
  *fact  = FACTZ;
}

void zgsrot_(iwtrot)
     int *iwtrot;
{

  /* set frame rotation flag */

  iwtroz = *iwtrot;
}

/*---------------------internal function ------------------*/

void zgbmcv_(nx, ny, bmline, bitmap)
     int *nx, *ny;
     char bmline[], bitmap[];
{
  static int n8 = 8;
  int n, nb;
  unsigned bx;

  nb = *nx * *ny / n8;
  for (n = 0; n < nb; n++){
    sscanf(&bmline[2*n], "%2x", &bx);
    bitmap[n] = bx;
  }
  bitmap[nb] = '\0';
}
